#include <botan/dsa.h>
#include <botan/rsa.h>
#include <botan/nr.h>
#include <botan/dh.h>
#include <botan/rw.h>
#include <botan/elgamal.h>

#include <botan/look_pk.h>
#include <botan/rng.h>
using namespace Botan;

#include <iostream>
#include <fstream>
#include <ctime>
#include <string>

#define DEBUG 0

BigInt to_bigint(const std::string&);

RSA_PrivateKey load_rsa_key(const std::string&);
RW_PrivateKey  load_rw_key(const std::string&);

void bench_enc(PK_Encryptor*, const std::string&, double, bool);
void bench_dec(PK_Encryptor*, PK_Decryptor*, const std::string&, double, bool);
void bench_sig(PK_Signer*, const std::string&, double, bool);
void bench_ver(PK_Signer*, PK_Verifier*, const std::string&, double, bool);
void bench_kas(PK_Key_Agreement_Scheme*, const std::string&, double, bool);

void bench_rsa(RSA_PrivateKey& key, const std::string keybits,
               double seconds, bool html)
   {
   bench_enc(get_pk_encryptor(key, "Raw"),
             "RSA-" + keybits, seconds, html);
   bench_dec(get_pk_encryptor(key, "Raw"),
             get_pk_decryptor(key, "Raw"),
             "RSA-" + keybits, seconds, html);
   }

void bench_rw(RW_PrivateKey& key, const std::string keybits,
              double seconds, bool html)
   {
   bench_ver(get_pk_signer(key, "EMSA2(SHA-1)"),
             get_pk_verifier(key, "EMSA2(SHA-1)"),
             "RW-" + keybits, seconds, html);
   bench_sig(get_pk_signer(key, "EMSA2(SHA-1)"),
             "RW-" + keybits, seconds, html);
   }

void bench_dsa(DSA_PrivateKey& key, const std::string keybits,
               double seconds, bool html)
   {
   bench_ver(get_pk_signer(key, "EMSA1(SHA-1)"),
             get_pk_verifier(key, "EMSA1(SHA-1)"),
             "DSA-" + keybits, seconds, html);
   bench_sig(get_pk_signer(key, "EMSA1(SHA-1)"),
             "DSA-" + keybits, seconds, html);
   }

void bench_nr(NR_PrivateKey& key, const std::string keybits,
              double seconds, bool html)
   {
   bench_ver(get_pk_signer(key, "EMSA1(SHA-1)"),
             get_pk_verifier(key, "EMSA1(SHA-1)"),
             "NR-" + keybits, seconds, html);
   bench_sig(get_pk_signer(key, "EMSA1(SHA-1)"),
             "NR-" + keybits, seconds, html);
   }

void bench_elg(ElGamal_PrivateKey& key, const std::string keybits,
               double seconds, bool html)
   {
   bench_enc(get_pk_encryptor(key, "Raw"),
             "ELG-" + keybits, seconds, html);
   bench_dec(get_pk_encryptor(key, "Raw"),
             get_pk_decryptor(key, "Raw"),
             "ELG-" + keybits, seconds, html);
   }

void bench_dh(DH_PrivateKey& key, const std::string keybits,
              double seconds, bool html)
   {
   bench_kas(get_pk_kas(key, "Raw"),
             "DH-" + keybits, seconds, html);
   }

void bench_pk(const std::string& algo, bool html, double seconds)
   {
   /*
     There is some strangeness going on here. It looks like algorithms
     at the end take some kind of pentalty. For example, running the RW tests
     first got a result of:
         RW-1024: 148.14 ms / private operation
     but running them last output:
         RW-1024: 363.54 ms / private operation

     I think it's from memory fragmentation in the allocators, but I'm
     not really sure. Need to investigate.

     Until then, I've basically ordered the tests in order of most important
     algorithms (RSA, DSA) to least important (NR, RW).
   */
   try {

   if(algo == "All" || algo == "RSA")
      {
      RSA_PrivateKey rsa512 = load_rsa_key("checks/keys/rsa512.key");
      bench_rsa(rsa512,  "512",  seconds, html);

      RSA_PrivateKey rsa1024 = load_rsa_key("checks/keys/rsa1024.key");
      bench_rsa(rsa1024, "1024", seconds, html);

      RSA_PrivateKey rsa2048 = load_rsa_key("checks/keys/rsa2048.key");
      bench_rsa(rsa2048, "2048", seconds, html);
      }
   if(algo == "All" || algo == "DSA")
      {
      DSA_PrivateKey dsa512(get_dl_group("DSA-512"));
      DSA_PrivateKey dsa768(get_dl_group("DSA-768"));
      DSA_PrivateKey dsa1024(get_dl_group("DSA-1024"));
      bench_dsa(dsa512,  "512",  seconds, html);
      bench_dsa(dsa768,  "768",  seconds, html);
      bench_dsa(dsa1024, "1024", seconds, html);
      }
   if(algo == "All" || algo == "DH")
      {
      DH_PrivateKey dh768(get_dl_group("IETF-768"));
      DH_PrivateKey dh1024(get_dl_group("IETF-1024"));
      DH_PrivateKey dh1536(get_dl_group("IETF-1536"));
      DH_PrivateKey dh2048(get_dl_group("IETF-2048"));
      bench_dh(dh768,  "768",  seconds, html);
      bench_dh(dh1024, "1024", seconds, html);
      bench_dh(dh1536, "1536", seconds, html);
      bench_dh(dh2048, "2048", seconds, html);
      }
   if(algo == "All" || algo == "ELG" || algo == "ElGamal")
      {
      ElGamal_PrivateKey elg768(get_dl_group("IETF-768"));
      ElGamal_PrivateKey elg1024(get_dl_group("IETF-1024"));
      ElGamal_PrivateKey elg1536(get_dl_group("IETF-1536"));
      ElGamal_PrivateKey elg2048(get_dl_group("IETF-2048"));
      bench_elg(elg768,  "768",  seconds, html);
      bench_elg(elg1024, "1024", seconds, html);
      bench_elg(elg1536, "1536", seconds, html);
      bench_elg(elg2048, "2048", seconds, html);
      }
   if(algo == "All" || algo == "NR")
      {
      NR_PrivateKey nr512(get_dl_group("DSA-512"));
      NR_PrivateKey nr768(get_dl_group("DSA-768"));
      NR_PrivateKey nr1024(get_dl_group("DSA-1024"));
      bench_nr(nr512,  "512",  seconds, html);
      bench_nr(nr768,  "768",  seconds, html);
      bench_nr(nr1024, "1024", seconds, html);
      }
   if(algo == "All" || algo == "RW")
      {
      RW_PrivateKey rw512 = load_rw_key("checks/keys/rw512.key");
      RW_PrivateKey rw1024 = load_rw_key("checks/keys/rw1024.key");
      bench_rw(rw512,  "512",  seconds, html);
      bench_rw(rw1024, "1024", seconds, html);
      }
   }
   catch(Botan::Exception& e)
      {
      std::cout << "Exception caught: " << e.what() << std::endl;
      return;
      }
   catch(std::exception& e)
      {
      std::cout << "Standard library exception caught: "
                << e.what() << std::endl;
      return;
      }
   catch(...)
      {
      std::cout << "Unknown exception caught." << std::endl;
      return;
      }

   }

void print_result(bool, u32bit runs, u32bit clocks_used,
                  const std::string& algo_name, const std::string& op)
   {
   std::cout << algo_name << ": ";

   double seconds = (double)clocks_used / CLOCKS_PER_SEC;
   double mseconds_per_run = 1000 * (seconds / runs);

   std::cout.setf(std::ios::fixed, std::ios::floatfield);
   std::cout.precision(2);
   std::cout << mseconds_per_run << " ms / " << op << std::endl;
   }

void bench_enc(PK_Encryptor* enc, const std::string& algo_name,
               double seconds, bool html)
   {
   static const u32bit MSG_SIZE = 16;
   byte msg[MSG_SIZE];

   u32bit runs = 0;
   std::clock_t start = std::clock(), clocks_used = 0;

   while(clocks_used < seconds * CLOCKS_PER_SEC)
      {
      runs++;
      Global_RNG::randomize(msg, MSG_SIZE);
      enc->encrypt(msg, MSG_SIZE);
      clocks_used = std::clock() - start;
      }

   print_result(html, runs, clocks_used, algo_name, "public operation");
   }

void bench_dec(PK_Encryptor* enc, PK_Decryptor* dec,
               const std::string& algo_name,
               double seconds, bool html)
   {
   static const u32bit MSG_SIZE = 16;
   byte msg[MSG_SIZE];
   Global_RNG::randomize(msg, MSG_SIZE);

   u32bit runs = 0;
   std::clock_t start = std::clock(), clocks_used = 0;

   SecureVector<byte> encrypted_msg = enc->encrypt(msg, MSG_SIZE);

   while(clocks_used < seconds * CLOCKS_PER_SEC)
      {
      runs++;
      dec->decrypt(encrypted_msg, encrypted_msg.size());
      clocks_used = std::clock() - start;
      }

   print_result(html, runs, clocks_used, algo_name, "private operation");
   }

void bench_sig(PK_Signer* sig, const std::string& algo_name,
               double seconds, bool html)
   {
   static const u32bit MSG_SIZE = 16;
   byte msg[MSG_SIZE];

   u32bit runs = 0;
   std::clock_t start = std::clock(), clocks_used = 0;

   while(clocks_used < seconds * CLOCKS_PER_SEC)
      {
      runs++;
      Global_RNG::randomize(msg, MSG_SIZE);
      sig->update(msg, MSG_SIZE);
      sig->signature();
      clocks_used = std::clock() - start;
      }

   print_result(html, runs, clocks_used, algo_name, "private operation");
   }

void bench_ver(PK_Signer* sig, PK_Verifier* ver,
               const std::string& algo_name,
               double seconds, bool html)
   {
   static const u32bit MSG_SIZE = 16;
   byte msg[MSG_SIZE];
   Global_RNG::randomize(msg, MSG_SIZE);

   sig->update(msg, MSG_SIZE);
   SecureVector<byte> signature = sig->signature();

   u32bit runs = 0;
   std::clock_t start = std::clock(), clocks_used = 0;

   while(clocks_used < seconds * CLOCKS_PER_SEC)
      {
      runs++;
      ver->update(msg, MSG_SIZE);
      ver->valid_signature(signature, signature.size());
      clocks_used = std::clock() - start;
      }

   print_result(html, runs, clocks_used, algo_name, "public operation");
   }

void bench_kas(PK_Key_Agreement_Scheme* kas, const std::string& algo_name,
               double seconds, bool html)
   {
   /* 128 bits: should always be considered valid (what about ECC?) */
   static const u32bit REMOTE_KEY_SIZE = 16;
   byte key[REMOTE_KEY_SIZE];

   u32bit runs = 0;
   std::clock_t start = std::clock(), clocks_used = 0;

   while(clocks_used < seconds * CLOCKS_PER_SEC)
      {
      runs++;
      Global_RNG::randomize(key, REMOTE_KEY_SIZE);
      kas->derive_key(key, REMOTE_KEY_SIZE, 0);
      clocks_used = std::clock() - start;
      }

   print_result(html, runs, clocks_used, algo_name, "key agreement");
   }

/*************************************************
* Key loading procedures                         *
*************************************************/
RSA_PrivateKey load_rsa_key(const std::string& file)
   {
   std::ifstream keyfile(file.c_str());
   if(!keyfile)
      throw Exception("Couldn't open the RSA key file " + file);

   std::string e, p, q;

   std::getline(keyfile, e);
   std::getline(keyfile, p);
   std::getline(keyfile, q);

   RSA_PrivateKey key(to_bigint(p), to_bigint(q), to_bigint(e));

#if DEBUG
   if(!key.check_params())
      throw Exception("Bad key from " + file);
#endif
   return key;
   }

RW_PrivateKey load_rw_key(const std::string& file)
   {
   std::ifstream keyfile(file.c_str());
   if(!keyfile)
      throw Exception("Couldn't open the RW key file " + file);

   std::string e, p, q;

   std::getline(keyfile, e);
   std::getline(keyfile, p);
   std::getline(keyfile, q);

   RW_PrivateKey key(to_bigint(p), to_bigint(q), to_bigint(e));

#if DEBUG
   if(!key.check_params())
      throw Exception("Bad key from " + file);
#endif
   return key;
   }
